import dataclasses
import typing
import cairo
import math
import gi

gi.require_version("PangoCairo", "1.0")
import gi.repository

import momapy.drawing
import momapy.geometry
import momapy.builder
import momapy.rendering.core


@dataclasses.dataclass
class CairoRenderer(momapy.rendering.core.Renderer):
    formats: typing.ClassVar[list[str]] = ["pdf", "svg", "png", "ps"]
    _de_class_func_mapping: typing.ClassVar[dict] = {
        momapy.drawing.Group: "_render_group",
        momapy.drawing.Path: "_render_path",
        momapy.drawing.Text: "_render_text",
        momapy.drawing.Ellipse: "_render_ellipse",
        momapy.drawing.Rectangle: "_render_rectangle",
    }
    _pa_class_func_mapping: typing.ClassVar[dict] = {
        momapy.drawing.MoveTo: "_render_move_to",
        momapy.drawing.LineTo: "_render_line_to",
        momapy.drawing.CurveTo: "_render_curve_to",
        momapy.drawing.QuadraticCurveTo: "_render_quadratic_curve_to",
        momapy.drawing.Close: "_render_close",
        momapy.drawing.EllipticalArc: "_render_elliptical_arc",
    }
    _tr_class_func_mapping: typing.ClassVar[dict] = {
        momapy.geometry.Translation: "_add_translation",
        momapy.geometry.Rotation: "_add_rotation",
        momapy.geometry.Scaling: "_add_scaling",
        momapy.geometry.MatrixTransformation: "_add_matrix_transformation",
    }
    context: cairo.Context
    config: dict = dataclasses.field(default_factory=dict)
    _pango_font_descriptions: dict = dataclasses.field(default_factory=dict)

    @classmethod
    def from_file(cls, output_file, width, height, format_, config=None):
        def _make_surface(output_file, width, height, format_):
            if format_ == "pdf":
                return cairo.PDFSurface(output_file, width, height)
            elif format_ == "ps":
                return cairo.PSSurface(output_file, width, height)
            elif format_ == "svg":
                return cairo.SVGSurface(output_file, width, height)
            elif format_ == "png":
                return cairo.ImageSurface(
                    cairo.FORMAT_ARGB32, int(width), int(height)
                )

        if config is None:
            config = {}
        config["output_file"] = output_file
        config["format"] = format_
        config["width"] = width
        config["height"] = height
        surface = _make_surface(output_file, width, height, format_)
        return cls.from_surface(surface, config)

    @classmethod
    def from_surface(cls, surface, config=None):
        if config is None:
            config = {}
        config["surface"] = surface
        context = cairo.Context(surface)
        return cls(context, config)

    def begin_session(self):
        self._states = []
        self._stroke = None
        self._fill = self.default_fill
        self._stroke_width = self.default_stroke_width

    def end_session(self):
        surface = self.context.get_target()
        format_ = self.config.get("format")
        if format_ == "png":
            surface.write_to_png(self.config["output_file"])
        surface.finish()
        surface.flush()

    def new_page(self, width, height):
        format_ = self.config.get("format")
        if format_ == "pdf" or format_ == "ps":
            self.context.show_page()
            surface = self.context.get_target()
            surface.set_size(width, height)

    def render_map(self, map_):
        self.render_layout_element(map_.layout)

    def render_layout_element(self, layout_element):
        drawing_elements = layout_element.drawing_elements()
        for drawing_element in drawing_elements:
            self.render_drawing_element(drawing_element)

    def render_drawing_element(self, drawing_element):
        self._save()
        self._set_state_from_drawing_element(drawing_element)
        self._add_transform_from_drawing_element(drawing_element)
        class_ = type(drawing_element)
        if issubclass(class_, momapy.builder.Builder):
            class_ = class_._cls_to_build
        de_func = getattr(self, self._de_class_func_mapping[class_])
        de_func(drawing_element)
        self._restore()

    def _save(self):
        state = {
            "stroke": self._stroke,
            "fill": self._fill,
            "stroke_width": self._stroke_width,
        }
        self._states.append(state)
        self.context.save()

    def _restore(self):
        state = self._states.pop()
        self._set_state(state)
        self.context.restore()
        self._set_new_path()  # context.restore() does not forget the current path

    def _set_state(self, state):
        for key in state:
            setattr(self, f"_{key}", state[key])

    def _set_state_from_drawing_element(self, drawing_element):
        state = self._get_state_from_drawing_element(drawing_element)
        self._set_state(state)

    def _get_state_from_drawing_element(self, drawing_element):
        state = {}
        if drawing_element.stroke is momapy.drawing.NoneValue:
            state["stroke"] = None
        elif drawing_element.stroke is not None:
            state["stroke"] = drawing_element.stroke
        if drawing_element.fill is momapy.drawing.NoneValue:
            state["fill"] = None
        elif drawing_element.fill is not None:
            state["fill"] = drawing_element.fill
        if (
            drawing_element.stroke_width is not None
        ):  # not sure, need to check svg spec
            state["stroke_width"] = drawing_element.stroke_width
        return state

    def _set_new_path(self):
        self.context.new_path()

    def _stroke_and_fill(self):
        if self._fill is not None:
            self.context.set_source_rgba(*self._fill.to_rgba(rgba_range=(0, 1)))
            if self._stroke is not None:
                self.context.fill_preserve()
            else:
                self.context.fill()
        if self._stroke is not None:
            self.context.set_line_width(self._stroke_width)
            self.context.set_source_rgba(
                *self._stroke.to_rgba(rgba_range=(0, 1))
            )
            self.context.stroke()

    def _render_path_action(self, path_action):
        class_ = type(path_action)
        if issubclass(class_, momapy.builder.Builder):
            class_ = class_._cls_to_build
        pa_func = getattr(self, self._pa_class_func_mapping[class_])
        pa_func(path_action)

    def _add_transform_from_drawing_element(self, drawing_element):
        if drawing_element.transform is not None:
            for transformation in drawing_element.transform:
                self._add_transformation(transformation)

    def _add_transformation(self, transformation):
        class_ = type(transformation)
        if issubclass(class_, momapy.builder.Builder):
            class_ = class_._cls_to_build
        tr_func = getattr(
            self, self._tr_class_func_mapping[type(transformation)]
        )
        return tr_func(transformation)

    def _render_group(self, group):
        for drawing_element in group.elements:
            self.render_drawing_element(drawing_element)

    def _render_path(self, path):
        for path_action in path.actions:
            self._render_path_action(path_action)
        self._stroke_and_fill()

    def _render_text(self, text):
        pango_layout = gi.repository.PangoCairo.create_layout(self.context)
        pango_font_description = self._pango_font_descriptions.get(
            (
                text.font_family,
                text.font_size,
            )
        )
        if pango_font_description is None:
            pango_font_description = gi.repository.Pango.FontDescription()
            pango_font_description.set_family(text.font_family)
            pango_font_description.set_absolute_size(
                gi.repository.Pango.units_from_double(text.font_size)
            )
            self._pango_font_descriptions[
                (text.font_family, text.font_size)
            ] = pango_font_description
        pango_layout.set_font_description(pango_font_description)
        pango_layout.set_text(text.text)
        pos = pango_layout.index_to_pos(0)
        x = gi.repository.Pango.units_to_double(pos.x)
        pango_layout_iter = pango_layout.get_iter()
        y = round(
            gi.repository.Pango.units_to_double(
                pango_layout_iter.get_baseline()
            )
        )
        tx = text.x - x
        ty = text.y - y
        self.context.translate(tx, ty)
        self.context.set_source_rgba(*self._fill.to_rgba(rgba_range=(0, 1)))
        gi.repository.PangoCairo.show_layout(self.context, pango_layout)

    def _render_ellipse(self, ellipse):
        self.context.save()
        self.context.translate(ellipse.x, ellipse.y)
        self.context.scale(ellipse.rx, ellipse.ry)
        self.context.arc(0, 0, 1, 0, 2 * math.pi)
        self.context.close_path()
        self.context.restore()
        self._stroke_and_fill()

    def _render_rectangle(self, rectangle):
        actions.append(rectangle.to_path())
        self._render_path(path)

    def _render_MoveTo(self, move_to):
        self.context.MoveTo(move_to.x, move_to.y)

    def _render_LineTo(self, line_to):
        self.context.LineTo(line_to.x, line_to.y)

    def _render_CurveTo(self, curve_to):
        self.context.CurveTo(
            curve_to.control_point1.x,
            curve_to.control_point2.y,
            curve_to.control_point2.x,
            curve_to.control_point2.y,
            curve_to.x,
            curve_to.y,
        )

    def _render_QuadraticCurveTo(self, quadratic_curve_to):
        cairo_current_point = self.context.get_current_point()
        current_point = momapy.geometry.Point(
            cairo_current_point[0], cairo_current_point[1]
        )
        curve_to = quadratic_curve_to.to_cubic(current_point)
        self._render_CurveTo(curve_to)

    def _render_ClosePath(self, close):
        self.context.close_path()

    def _render_EllipticalArc(self, elliptical_arc):
        obj = momapy.geometry.EllipticalArc(
            momapy.geometry.Point(
                self.context.get_current_point()[0],
                self.context.get_current_point()[1],
            ),
            elliptical_arc.point,
            elliptical_arc.rx,
            elliptical_arc.ry,
            elliptical_arc.x_axis_rotation,
            elliptical_arc.arc_flag,
            elliptical_arc.sweep_flag,
        )
        arc, transformation = obj.to_arc_and_transformation()
        self.context.save()
        self._add_transformation(transformation)
        self.context.arc(
            arc.point.x, arc.point.y, arc.radius, arc.start_angle, arc.end_angle
        )
        self.context.restore()

    def _add_translation(self, translation):
        self.context.translate(translation.tx, translation.ty)

    def _add_rotation(self, rotation):
        point = rotation.point
        if point is not None:
            self.context.translate(point.x, point.y)
            self.context.rotate(rotation.angle)
            self.context.translate(-point.x, -point.y)
        else:
            self.context.rotate(rotation.angle)

    def _add_scaling(self, scaling):
        self.context.scale(scaling.sx, scaling.sy)

    def _add_matrix_transformation(self, matrix_transformation):
        m = cairo.Matrix(
            xx=matrix_transformation.m[0][0],
            yx=matrix_transformation.m[1][0],
            xy=matrix_transformation.m[0][1],
            yy=matrix_transformation.m[1][1],
            x0=matrix_transformation.m[0][2],
            y0=matrix_transformation.m[1][2],
        )
        self.context.transform(m)


momapy.rendering.core.register_renderer("cairo", CairoRenderer)



"""Classes for rendering in the SVG format"""

import dataclasses
import typing
import math
import xml.sax.saxutils

import momapy.drawing
import momapy.geometry
import momapy.rendering.core


@dataclasses.dataclass
class SVGElement(object):
    """Class for SVG elements"""

    name: str
    value: typing.Optional[str] = None
    attributes: dict = dataclasses.field(default_factory=dict)
    elements: list["SVGElement"] = dataclasses.field(default_factory=list)

    def to_string(self, indent: int = 0):
        """Return the SVG string representing the element"""
        s_indent = "\t" * indent
        s_value = f"{s_indent}{self.value}\n" if self.value is not None else ""
        if self.attributes:
            l_s_attributes = []
            for attr_name, attr_value in self.attributes.items():
                s_attr_name = attr_name
                s_attr_value = f'"{attr_value}"'
                s_attribute = f"{s_attr_name}={s_attr_value}"
                l_s_attributes.append(s_attribute)
            s_attributes = f" {' '.join(l_s_attributes)}"
        else:
            s_attributes = ""
        if self.elements:
            s_elements = "\n".join(
                [child.to_string(indent + 1) for child in self.elements]
            )
            s_elements += "\n"
        else:
            s_elements = ""
        return f"{s_indent}<{self.name}{s_attributes}>\n{s_value}{s_elements}{s_indent}</{self.name}>"

    def __str__(self):
        return self.to_string()

    def add_element(self, element):
        """Add an sub-element to the SVG element"""
        self.elements.append(element)


@dataclasses.dataclass
class SVGNativeRenderer(momapy.rendering.core.Renderer):
    """Class for SVG native renderers"""

    formats: typing.ClassVar[list[str]] = ["svg"]
    _de_class_func_mapping: typing.ClassVar[dict] = {
        momapy.drawing.Group: "_make_group_element",
        momapy.drawing.Path: "_make_path_element",
        momapy.drawing.Text: "_make_text_element",
        momapy.drawing.Ellipse: "_make_ellipse_element",
        momapy.drawing.Rectangle: "_make_rectangle_element",
    }
    _pa_class_func_mapping: typing.ClassVar[dict] = {
        momapy.drawing.MoveTo: "_make_move_to_value",
        momapy.drawing.LineTo: "_make_line_to_value",
        momapy.drawing.CurveTo: "_make_curve_to_value",
        momapy.drawing.QuadraticCurveTo: "_make_quadratic_curve_to_value",
        momapy.drawing.ClosePath: "_make_close_value",
        momapy.drawing.EllipticalArc: "_make_elliptical_arc_value",
    }
    _tr_class_func_mapping: typing.ClassVar[dict] = {
        momapy.geometry.Translation: "_make_translation_value",
        momapy.geometry.Rotation: "_make_rotation_value",
        momapy.geometry.Scaling: "_make_scaling_value",
        momapy.geometry.MatrixTransformation: "_make_matrix_transformation_value",
    }
    _fe_class_func_mapping: typing.ClassVar[dict] = {
        momapy.drawing.DropShadowEffect: "_make_drop_shadow_effect_element",
        momapy.drawing.CompositeEffect: "_make_composite_effect_element",
        momapy.drawing.GaussianBlurEffect: "_make_gaussian_blur_effect_element",
        momapy.drawing.OffsetEffect: "_make_offset_effect_element",
        momapy.drawing.FloodEffect: "_make_flood_effect_element",
    }
    _fe_composite_comp_op_value_mapping: typing.ClassVar[dict] = {
        momapy.drawing.CompositionOperator.OVER: "over",
        momapy.drawing.CompositionOperator.IN: "in",
        momapy.drawing.CompositionOperator.OUT: "out",
        momapy.drawing.CompositionOperator.ATOP: "atop",
        momapy.drawing.CompositionOperator.XOR: "xor",
        momapy.drawing.CompositionOperator.LIGHTER: "lighter",
        momapy.drawing.CompositionOperator.ARTIHMETIC: "arithmetic",
    }
    _fe_gaussian_blur_edgemode_value_mapping: typing.ClassVar[dict] = {
        momapy.drawing.EdgeMode.WRAP: "wrap",
        momapy.drawing.EdgeMode.DUPLICATE: "duplicate",
        momapy.drawing.NoneValue: "none",
    }
    _fe_filter_unit_value_mapping: typing.ClassVar[dict] = {
        momapy.drawing.FilterUnits.USER_SPACE_ON_USE: "UserSpaceOnUse",
        momapy.drawing.FilterUnits.OBJECT_BOUNDING_BOX: "objectBoundingBox",
    }
    _fe_input_value_mapping: typing.ClassVar[dict] = {
        momapy.drawing.FilterEffectInput.SOURCE_GRAPHIC: "SourceGraphic",
        momapy.drawing.FilterEffectInput.SOURCE_ALPHA: "SourceAlpha",
        momapy.drawing.FilterEffectInput.BACKGROUND_IMAGE: "BackgroundImage",
        momapy.drawing.FilterEffectInput.BACKGROUND_ALPHA: "BackgroundAlpha",
        momapy.drawing.FilterEffectInput.FILL_PAINT: "FillPaint",
        momapy.drawing.FilterEffectInput.STROKE_PAINT: "StrokePaint",
    }
    _te_font_style_value_mapping: typing.ClassVar[dict] = {
        momapy.drawing.FontStyle.NORMAL: "normal",
        momapy.drawing.FontStyle.ITALIC: "italic",
        momapy.drawing.FontStyle.OBLIQUE: "oblique",
    }
    _te_font_weight_value_mapping: typing.ClassVar[dict] = {
        momapy.drawing.FontWeight.NORMAL: "normal",
        momapy.drawing.FontWeight.BOLD: "bold",
        momapy.drawing.FontWeight.BOLDER: "bolder",
        momapy.drawing.FontWeight.LIGHTER: "lighter",
    }
    _te_text_anchor_value_mapping: typing.ClassVar[dict] = {
        momapy.drawing.TextAnchor.START: "start",
        momapy.drawing.TextAnchor.MIDDLE: "middle",
        momapy.drawing.TextAnchor.END: "end",
    }
    _de_fill_rule_value_mapping: typing.ClassVar[dict] = {
        momapy.drawing.FillRule.NONZERO: "nonzero",
        momapy.drawing.FillRule.EVENODD: "evenodd",
    }

    svg: SVGElement
    config: dict = dataclasses.field(default_factory=dict)
    _filter_elements: list[SVGElement] = dataclasses.field(default_factory=list)

    @classmethod
    def from_file(cls, output_file, width, height, format_, config=None):
        if config is None:
            config = {}
        config["output_file"] = output_file
        config["width"] = width
        config["height"] = height
        config["format"] = format_
        svg = SVGElement(
            name="svg",
            attributes={
                "xmlns": "http://www.w3.org/2000/svg",
                "viewBox": f"0 0 {width} {height}",
            },
        )
        return cls(svg=svg, config=config)

    def begin_session(self):
        pass

    def end_session(self):
        if self._filter_elements:
            defs = SVGElement(name="defs", elements=self._filter_elements)
            self.svg.add_element(defs)
        if self.config.get("output_file") is not None:
            with open(self.config["output_file"], "w", encoding="utf-8") as f:
                f.write(str(self.svg))

    def new_page(self, width, height):
        pass

    def render_map(self, map_):
        self.render_layout_element(map_.layout)

    def render_layout_element(self, layout_element):
        drawing_elements = layout_element.drawing_elements()
        for drawing_element in drawing_elements:
            self.render_drawing_element(drawing_element)

    def render_drawing_element(self, drawing_element):
        element = self._make_drawing_element_element(drawing_element)
        self.svg.add_element(element)

    def _make_color_value(self, color):
        return f"rgb({color.red}, {color.green}, {color.blue})"

    def _make_opacity_value(self, color):
        return str(color.alpha)

    def _make_transform_value(self, transform):
        value = " ".join(
            [
                self._make_transformation_value(transformation)
                for transformation in transform
            ]
        )
        return value

    def _make_transformation_value(self, transformation):
        class_ = type(transformation)
        if issubclass(class_, momapy.builder.Builder):
            class_ = class_._cls_to_build
        tr_func = getattr(self, self._tr_class_func_mapping[class_])
        return tr_func(transformation)

    def _make_drawing_element_element_id_class_atrributes(self, drawing_element):
        attributes = {}
        id_ = getattr(drawing_element, "id_", None)
        if id_ is not None:
            attributes["id"] = id_
        class_ = getattr(drawing_element, "class_", None)
        if class_ is not None:
            attributes["class"] = class_
        return attributes

    def _make_drawing_element_presentation_attributes(self, drawing_element):
        attributes = {}
        for attr_name in momapy.drawing.PRESENTATION_ATTRIBUTES:
            attr_value = getattr(drawing_element, attr_name)
            if attr_value is not None:
                if attr_value == momapy.drawing.NoneValue:
                    attr_value = "none"
                else:
                    if attr_name == "stroke" or attr_name == "fill":
                        opacity_value = self._make_opacity_value(attr_value)
                        attributes[f"{attr_name}-opacity"] = opacity_value
                        attr_value = self._make_color_value(attr_value)
                    elif attr_name == "transform":
                        attr_value = self._make_transform_value(attr_value)
                    elif attr_name == "filter":
                        filter_element = self._make_filter_element(attr_value)
                        if filter_element not in self._filter_elements:
                            self._filter_elements.append(filter_element)
                        attr_value = f"url(#{attr_value.id_})"
                    elif attr_name == "font_style":
                        attr_value = self._te_font_style_value_mapping[attr_value]
                    elif attr_name == "font_weight":
                        if isinstance(attr_value, momapy.drawing.FontWeight):
                            attr_value = self._te_font_weight_value_mapping[attr_value]
                    elif attr_name == "text_anchor":
                        attr_value = self._te_text_anchor_value_mapping[attr_value]
                    elif attr_name == "fill_rule":
                        attr_value = self._de_fill_rule_value_mapping[attr_value]
                    elif attr_name == "stroke_dasharray":
                        attr_value = " ".join(
                            [
                                str(attr_value_element)
                                for attr_value_element in attr_value
                            ]
                        )
                attr_name = attr_name.replace("_", "-")
                attributes[attr_name] = attr_value
        return attributes

    def _make_filter_element(self, filter_):
        name = "filter"
        attributes = {}
        attributes["id"] = filter_.id_
        attributes["filterUnits"] = self._fe_filter_unit_value_mapping[
            filter_.filter_units
        ]
        attributes["x"] = filter_.x
        attributes["y"] = filter_.y
        attributes["width"] = filter_.width
        attributes["height"] = filter_.height
        subelements = []
        for filter_effect in filter_.effects:
            subelement = self._make_filter_effect_element(filter_effect)
            subelements.append(subelement)
        element = SVGElement(name=name, attributes=attributes, elements=subelements)
        return element

    def _make_filter_effect_element(self, filter_effect):
        class_ = type(filter_effect)
        if issubclass(class_, momapy.builder.Builder):
            class_ = class_._cls_to_build
        fe_func = getattr(self, self._fe_class_func_mapping[class_])
        element = fe_func(filter_effect)
        return element

    def _make_drop_shadow_effect_element(self, filter_effect):
        name = "feDropShadow"
        attributes = {}
        attributes["dx"] = filter_effect.dx
        attributes["dy"] = filter_effect.dy
        attributes["stdDeviation"] = filter_effect.std_deviation
        attributes["flood-opacity"] = filter_effect.flood_opacity
        attributes["flood-color"] = self._make_color_value(filter_effect.flood_color)
        if filter_effect.result is not None:
            attributes["result"] = filter_effect.result
        element = SVGElement(
            name=name,
            attributes=attributes,
        )
        return element

    def _make_composite_effect_element(self, filter_effect):
        name = "feComposite"
        attributes = {}
        if isinstance(filter_effect.in_, momapy.drawing.FilterEffectInput):
            attributes["in"] = self._fe_input_value_mapping[filter_effect.in_]
        else:
            attributes["in"] = filter_effect.in_
        if isinstance(filter_effect.in2, momapy.drawing.FilterEffectInput):
            attributes["in2"] = self._fe_input_value_mapping[filter_effect.in2]
        else:
            attributes["in2"] = filter_effect.in2
        attributes["operator"] = self._fe_composite_comp_op_value_mapping[
            filter_effect.operator
        ]
        if filter_effect.result is not None:
            attributes["result"] = filter_effect.result
        element = SVGElement(
            name=name,
            attributes=attributes,
        )
        return element

    def _make_flood_effect_element(self, filter_effect):
        name = "feFlood"
        attributes = {}
        attributes["flood-opacity"] = filter_effect.flood_opacity
        attributes["flood-color"] = self._make_color_value(filter_effect.flood_color)
        if filter_effect.result is not None:
            attributes["result"] = filter_effect.result
        element = SVGElement(
            name=name,
            attributes=attributes,
        )
        return element

    def _make_gaussian_blur_effect_element(self, filter_effect):
        name = "feGaussianBlur"
        attributes = {}
        if isinstance(filter_effect.in_, momapy.drawing.FilterEffectInput):
            attributes["in"] = self._fe_input_value_mapping[filter_effect.in_]
        else:
            attributes["in"] = filter_effect.in_
        attributes["stdDeviation"] = filter_effect.std_deviation
        if filter_effect.edge_mode is not None:
            attributes["edgeMode"] = self._fe_gaussian_blur_edgemode_value_mapping[
                filter_effect.edge_mode
            ]
        if filter_effect.result is not None:
            attributes["result"] = filter_effect.result
        element = SVGElement(
            name=name,
            attributes=attributes,
        )
        return element

    def _make_offset_effect_element(self, filter_effect):
        name = "feOffset"
        attributes = {}
        attributes["dx"] = filter_effect.dx
        attributes["dy"] = filter_effect.dy
        if filter_effect.result is not None:
            attributes["result"] = filter_effect.result
        element = SVGElement(
            name=name,
            attributes=attributes,
        )
        return element

    def _make_drawing_element_element(self, drawing_element):
        class_ = type(drawing_element)
        if issubclass(class_, momapy.builder.Builder):
            class_ = class_._cls_to_build
        de_func = getattr(self, self._de_class_func_mapping[class_])
        element = de_func(drawing_element)
        return element

    def _make_group_element(self, group):
        name = "g"
        presentation_attributes = self._make_drawing_element_presentation_attributes(
            group
        )
        id_class_attributes = self._make_drawing_element_element_id_class_atrributes(
            group
        )
        attributes = presentation_attributes | id_class_attributes
        subelements = []
        for drawing_element in group.elements:
            subelement = self._make_drawing_element_element(drawing_element)
            subelements.append(subelement)
        element = SVGElement(name=name, attributes=attributes, elements=subelements)
        return element

    def _make_path_action_value(self, path_action):
        class_ = type(path_action)
        if issubclass(class_, momapy.builder.Builder):
            class_ = class_._cls_to_build
        pa_func = getattr(self, self._pa_class_func_mapping[class_])
        value = pa_func(path_action)
        return value

    def _make_move_to_value(self, move_to):
        return f"M {move_to.x} {move_to.y}"

    def _make_line_to_value(self, line_to):
        return f"L {line_to.x} {line_to.y}"

    def _make_close_value(self, close):
        return "Z"

    def _make_elliptical_arc_value(self, elliptical_arc):
        return (
            f"A {elliptical_arc.rx} "
            f"{elliptical_arc.ry} "
            f"{elliptical_arc.x_axis_rotation} "
            f"{elliptical_arc.arc_flag} "
            f"{elliptical_arc.sweep_flag} "
            f"{elliptical_arc.x} "
            f"{elliptical_arc.y}"
        )

    def _make_quadratic_curve_to_value(self, quadratic_curve_to):
        return (
            f"Q {quadratic_curve_to.control_point.x} "
            f"{quadratic_curve_to.control_point.y} "
            f"{quadratic_curve_to.x} "
            f"{quadratic_curve_to.y}"
        )

    def _make_curve_to_value(self, curve_to):
        return (
            f"C {curve_to.control_point1.x} "
            f"{curve_to.control_point1.y} "
            f"{curve_to.control_point2.x} "
            f"{curve_to.control_point2.y} "
            f"{curve_to.x} "
            f"{curve_to.y}"
        )

    def _make_path_element(self, path):
        name = "path"
        presentation_attributes = self._make_drawing_element_presentation_attributes(
            path
        )
        id_class_attributes = self._make_drawing_element_element_id_class_atrributes(
            path
        )
        attributes = presentation_attributes | id_class_attributes
        d_value = " ".join(
            [self._make_path_action_value(path_action) for path_action in path.actions]
        )
        attributes["d"] = d_value
        element = SVGElement(name=name, attributes=attributes)
        return element

    def _make_text_element(self, text):
        name = "text"
        presentation_attributes = self._make_drawing_element_presentation_attributes(
            text
        )
        id_class_attributes = self._make_drawing_element_element_id_class_atrributes(
            text
        )
        attributes = presentation_attributes | id_class_attributes
        attributes["x"] = text.x
        attributes["y"] = text.y
        value = xml.sax.saxutils.escape(text.text)
        element = SVGElement(name=name, attributes=attributes, value=value)
        return element

    def _make_ellipse_element(self, ellipse):
        name = "ellipse"
        presentation_attributes = self._make_drawing_element_presentation_attributes(
            ellipse
        )
        id_class_attributes = self._make_drawing_element_element_id_class_atrributes(
            ellipse
        )
        attributes = presentation_attributes | id_class_attributes
        attributes["cx"] = ellipse.x
        attributes["cy"] = ellipse.y
        attributes["rx"] = ellipse.rx
        attributes["ry"] = ellipse.ry
        element = SVGElement(name=name, attributes=attributes)
        return element

    def _make_rectangle_element(self, rectangle):
        name = "rect"
        presentation_attributes = self._make_drawing_element_presentation_attributes(
            rectangle
        )
        id_class_attributes = self._make_drawing_element_element_id_class_atrributes(
            rectangle
        )
        attributes = presentation_attributes | id_class_attributes
        attributes["x"] = rectangle.x
        attributes["y"] = rectangle.y
        attributes["width"] = rectangle.width
        attributes["height"] = rectangle.height
        attributes["rx"] = rectangle.rx
        attributes["ry"] = rectangle.ry
        element = SVGElement(name=name, attributes=attributes)
        return element

    def _make_translation_value(self, translation):
        return f"translate({translation.tx} {translation.ty})"

    def _make_rotation_value(self, rotation):
        angle = math.degrees(rotation.angle)
        s_point = (
            f" {rotation.point.x} {rotation.point.y}"
            if rotation.point is not None
            else ""
        )
        value = f"rotate({angle}{s_point})"
        return value

    def _make_scaling_value(self, scaling):
        return f"scale({scaling.sx} {scaling.sy})"

    def _make_matrix_transformation_value(self, matrix_transformation):
        matrix_values = [
            matrix_transformation.m[i][j]
            for j in range(len(matrix_transformation.m[0]))
            for i in range(len(matrix_transformation.m) - 1)
        ]
        value = f"matrix({' '.join(matrix_values)})"
        return value


@dataclasses.dataclass
class SVGNativeCompatRenderer(SVGNativeRenderer):
    """Class for SVG native compat renderers"""

    def _make_filter_element(self, filter_):
        filter_ = filter_.to_compat()
        element = super()._make_filter_element(filter_)
        return element
